@@ -189,7 +189,7 @@ Client::Client(struct ev_loop *loop, uint32_t client_chosen_version,
     handshake_confirmed_(false),
     no_gso_{
 #ifdef UDP_SEGMENT
-      false
+      config.no_gso
 #else  // !defined(UDP_SEGMENT)
       true
 #endif // !defined(UDP_SEGMENT)
@@ -934,7 +934,14 @@ int Client::on_read(const Endpoint &ep) {
     .msg_control = msg_ctrl,
   };
 
-  for (;;) {
+  auto start = util::timestamp();
+
+  for (; pktcnt < MAX_RECV_PKTS;) {
+    if (util::recv_pkt_time_threshold_exceeded(
+          config.cc_algo == NGTCP2_CC_ALGO_BBR, start, pktcnt)) {
+      break;
+    }
+
     msg.msg_namelen = sizeof(su);
     msg.msg_controllen = sizeof(msg_ctrl);
 
@@ -995,10 +1002,6 @@ int Client::on_read(const Endpoint &ep) {
         break;
       }
     }
-
-    if (pktcnt >= 10) {
-      break;
-    }
   }
 
   if (should_exit()) {
@@ -1158,17 +1161,20 @@ int Client::write_streams() {
   size_t gso_size;
   auto ts = util::timestamp();
   auto txbuf = std::span{tx_.data};
+  auto buflen = util::clamp_buffer_size(conn_, txbuf.size(), config.gso_burst);
 
   ngtcp2_path_storage_zero(&ps);
 
-  auto nwrite =
-    ngtcp2_conn_write_aggregate_pkt(conn_, &ps.path, &pi, txbuf.data(),
-                                    txbuf.size(), &gso_size, ::write_pkt, ts);
+  auto nwrite = ngtcp2_conn_write_aggregate_pkt2(
+    conn_, &ps.path, &pi, txbuf.data(), buflen, &gso_size, ::write_pkt,
+    config.gso_burst, ts);
   if (nwrite < 0) {
     disconnect();
     return -1;
   }
 
+  ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+
   if (nwrite == 0) {
     return 0;
   }
@@ -1785,6 +1791,7 @@ int Client::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
     auto rv = nghttp3_conn_close_stream(httpconn_, stream_id, app_error_code);
     switch (rv) {
     case 0:
+      http_stream_close(stream_id, app_error_code);
       break;
     case NGHTTP3_ERR_STREAM_NOT_FOUND:
       // We have to handle the case when stream opened but no data is
@@ -1924,11 +1931,11 @@ int Client::submit_http_request(const Stream *stream) {
 
 int Client::recv_stream_data(uint32_t flags, int64_t stream_id,
                              std::span<const uint8_t> data) {
-  auto nconsumed =
-    nghttp3_conn_read_stream(httpconn_, stream_id, data.data(), data.size(),
-                             flags & NGTCP2_STREAM_DATA_FLAG_FIN);
+  auto nconsumed = nghttp3_conn_read_stream2(
+    httpconn_, stream_id, data.data(), data.size(),
+    flags & NGTCP2_STREAM_DATA_FLAG_FIN, ngtcp2_conn_get_timestamp(conn_));
   if (nconsumed < 0) {
-    std::cerr << "nghttp3_conn_read_stream: "
+    std::cerr << "nghttp3_conn_read_stream2: "
               << nghttp3_strerror(static_cast<int>(nconsumed)) << std::endl;
     ngtcp2_ccerr_set_application_error(
       &last_error_,
@@ -2149,37 +2156,26 @@ int Client::reset_stream(int64_t stream_id, uint64_t app_error_code) {
   return 0;
 }
 
-namespace {
-int http_stream_close(nghttp3_conn *conn, int64_t stream_id,
-                      uint64_t app_error_code, void *conn_user_data,
-                      void *stream_user_data) {
-  auto c = static_cast<Client *>(conn_user_data);
-  if (c->http_stream_close(stream_id, app_error_code) != 0) {
-    return NGHTTP3_ERR_CALLBACK_FAILURE;
-  }
-  return 0;
-}
-} // namespace
-
-int Client::http_stream_close(int64_t stream_id, uint64_t app_error_code) {
-  if (ngtcp2_is_bidi_stream(stream_id)) {
-    assert(ngtcp2_conn_is_local_stream(conn_, stream_id));
-
-    ++nstreams_closed_;
-  } else {
-    assert(!ngtcp2_conn_is_local_stream(conn_, stream_id));
-    ngtcp2_conn_extend_max_streams_uni(conn_, 1);
+void Client::http_stream_close(int64_t stream_id, uint64_t app_error_code) {
+  if (!ngtcp2_is_bidi_stream(stream_id)) {
+    return;
   }
 
-  if (auto it = streams_.find(stream_id); it != std::ranges::end(streams_)) {
-    if (!config.quiet) {
-      std::cerr << "HTTP stream " << stream_id << " closed with error code "
-                << app_error_code << std::endl;
-    }
-    streams_.erase(it);
+  assert(ngtcp2_conn_is_local_stream(conn_, stream_id));
+
+  ++nstreams_closed_;
+
+  auto it = streams_.find(stream_id);
+  if (it == std::ranges::end(streams_)) {
+    return;
   }
 
-  return 0;
+  if (!config.quiet) {
+    std::cerr << "HTTP stream " << stream_id << " closed with error code "
+              << app_error_code << std::endl;
+  }
+
+  streams_.erase(it);
 }
 
 namespace {
@@ -2226,7 +2222,6 @@ int Client::setup_httpconn() {
   }
 
   nghttp3_callbacks callbacks{
-    .stream_close = ::http_stream_close,
     .recv_data = ::http_recv_data,
     .deferred_consume = ::http_deferred_consume,
     .begin_headers = ::http_begin_headers,
@@ -2677,6 +2672,15 @@ Options:
               the handshake  fails with ech_required alert,  ECH retry
               configs,  if  provided by  server,  will  be written  to
               <PATH>.
+  --no-gso    Disables GSO.
+  --show-stat Print the connection statistics when the connection is
+              closed.
+  --gso-burst=<N>
+              The maximum number of packets  to aggregate for GSO.  If
+              GSO is disabled,  this is the maximum  number of packets
+              to send  per an event  loop in a single  connection.  It
+              defaults  to 0,  which means  it is  not limited  by the
+              configuration.
   -h, --help  Display this help and exit.
 
 ---
@@ -2762,6 +2766,9 @@ int main(int argc, char **argv) {
       {"initial-pkt-num", required_argument, &flag, 42},
       {"pmtud-probes", required_argument, &flag, 43},
       {"ech-config-list-file", required_argument, &flag, 44},
+      {"no-gso", no_argument, &flag, 45},
+      {"show-stat", no_argument, &flag, 46},
+      {"gso-burst", required_argument, &flag, 47},
       {},
     };
 
@@ -3092,9 +3099,9 @@ int main(int argc, char **argv) {
         if (auto n = util::parse_uint_iec(optarg); !n) {
           std::cerr << "max-udp-payload-size: invalid argument" << std::endl;
           exit(EXIT_FAILURE);
-        } else if (*n > 64_k) {
-          std::cerr << "max-udp-payload-size: must not exceed 65536"
-                    << std::endl;
+        } else if (*n > NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE) {
+          std::cerr << "max-udp-payload-size: must not exceed "
+                    << NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE << std::endl;
           exit(EXIT_FAILURE);
         } else if (*n == 0) {
           std::cerr << "max-udp-payload-size: must not be 0" << std::endl;
@@ -3212,10 +3219,12 @@ int main(int argc, char **argv) {
           if (auto n = util::parse_uint_iec(s); !n) {
             std::cerr << "pmtud-probes: invalid argument" << std::endl;
             exit(EXIT_FAILURE);
-          } else if (*n <= 1200 || *n >= 64_k) {
-            std::cerr
-              << "pmtud-probes: must be in range [1201, 65535], inclusive."
-              << std::endl;
+          } else if (*n <= NGTCP2_MAX_UDP_PAYLOAD_SIZE ||
+                     *n > NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE) {
+            std::cerr << "pmtud-probes: must be in range ["
+                      << NGTCP2_MAX_UDP_PAYLOAD_SIZE + 1 << ", "
+                      << NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE << "], inclusive."
+                      << std::endl;
             exit(EXIT_FAILURE);
           } else {
             config.pmtud_probes.push_back(static_cast<uint16_t>(*n));
@@ -3227,11 +3236,37 @@ int main(int argc, char **argv) {
         // --ech-config-list-file
         config.ech_config_list_file = optarg;
         break;
+      case 45:
+        // --no-gso
+        config.no_gso = true;
+        break;
+      case 46:
+        // --show-stat
+        config.show_stat = true;
+        break;
+      case 47: {
+        // --gso-burst
+        auto n = util::parse_uint(optarg);
+        if (!n) {
+          std::cerr << "gso-burst: invalid argument" << std::endl;
+          exit(EXIT_FAILURE);
+        }
+
+        if (*n > 64) {
+          std::cerr << "gso-burst: must be in range [0, 64], inclusive."
+                    << std::endl;
+          exit(EXIT_FAILURE);
+        }
+
+        config.gso_burst = static_cast<size_t>(*n);
+
+        break;
+      }
       }
       break;
     default:
       break;
-    };
+    }
   }
 
   if (argc - optind < 2) {
